// alineal.cpp
// Implementa la clase ALineal
#include<alineal.h>
#include<linalg.h>
#include<error.h>
#include<iostream>


// Funcin auxiliar de descomponeLU
static double sum(int i, int j, int ki, int kf, 
		  const RMatriz& lu){
    double sum = 0;
    for (int k = ki; k <= kf; k++){
	double bik = lu(i,k);
	double ckj = lu(k,j);
	sum += bik*ckj;
    }
    return sum;
}

//
// Clase ALineal
// Constructor
// Establece los interruptores necesarios para el
// funcionamiento del algoritmo
ALineal::ALineal(){

    // Aade interruptores (abiertos por defecto)

    // an no hemos creado objetos tipo matriz
    pCC.nuevoInterruptor("matriz"); 
    // an no hemos creado objetos tipo vector
    pCC.nuevoInterruptor("vector"); 
    // an no hemos factorizado la matriz
    pCC.nuevoInterruptor("lu"); 
    // an no hemos resuelto el sistema lineal 
    pCC.nuevoInterruptor("lineal"); 
    // an no hemos calculado inversa
    pCC.nuevoInterruptor("inversa"); 
    // an no hemos calculado el determinante
    pCC.nuevoInterruptor("determinante"); 
}

// Destructor
ALineal::~ALineal(){
    // si se crearon objetos de tipo matriz borralos
    if (pCC.interruptorCerrado("matriz")){
	delete pLU;
	delete pInv;
	delete pP;    
    }
    // si se crearon objetos de tipo vector borralos
    if (pCC.interruptorCerrado("vector")){
    delete pV;
    }
} 

// Estable la matriz del sistema lineal
void ALineal::eMatriz(const RMatriz& a)
{
    if (a.filas() != a.columnas())
	error("La descomposicin LU requiere una matriz cuadrada", 
	      CENOCUAD);

    // borra las estructuras antiguas si existen
    // en otro caso cierra el interruptor ya que vamos
    // a crear objetos de tipo matriz a continuacin

    if (pCC.interruptorCerrado("matriz")){
	delete pLU;
	delete pP;
	delete pInv;
    }else{
	pCC.cierraInterruptor("matriz");
    }

    // Crea los objetos tipo matriz necesarios
    // Dimensin
    pN = a.filas();
    // Crea un puntero a un objeto tipo matriz para LU
    pLU = new RMatriz(pN,pN);
    // Crea un puntero a un objeto tipo matriz para inversa
    pInv = new RMatriz(pN,pN);
    // Crea un puntero a un objeto tipo permutacin
    pP = new Permutacion(pN);
    // Copia a en lu
    *pLU = a;

    // Nueva matriz en la mquina. Por tanto hay que
    // factorizar de nuevo, resolver de nuevo el sistema
    // lineal si procede y calcular de nuevo la inversa
    // y el determinante. Abrimos los correspondientes
    // interruptores para indicarle al algoritmo que tiene
    // que ejectuarse de nuevo

    // an no hemos factorizado la matriz
    pCC.abreInterruptor("lu"); 
    // an no hemos resuelto el sistema lineal 
    pCC.abreInterruptor("lineal"); 
    // an no hemos calculado inversa
    pCC.abreInterruptor("inversa"); 
    // an no hemos calculado el determinante
    pCC.abreInterruptor("determinante"); 
}

// Establece el vector del sistema lineal
void ALineal::eVector(const RVector& v)
{

    // borra las estructuras antiguas si existen
    // en otro caso cierra el interruptor ya que vamos
    // a crear objetos de tipo vector a continuacin

    if (pCC.interruptorCerrado("vector")){
	delete pV;
    }else{
	pCC.cierraInterruptor("vector");
    }

    // crea un puntero a un vector de la dimensin de v
    pV = new RVector(v.dimension());
    // copia el vector v
    *pV = v;

    // Nuevo vector en la mquina. Por tanto hay que
    // resolver de nuevo el sistema
    // lineal 

    pCC.abreInterruptor("lineal"); 
}

// Establece el sistema lineal
void ALineal::eSistema(const RMatriz& a, const RVector& v)
{
    if (a.filas() != a.columnas())
	error("La descomposicin LU requiere una matriz cuadrada", 
	      CENOCUAD);

    if (a.filas() != v.dimension())
	error("LU: dime vector debe ser = a dim matriz", 
	      CEMALDIM);

    // borra las estructuras antiguas si existen
    // en otro caso cierra el interruptor ya que vamos
    // a crear objetos de tipo matriz a continuacin

    if (pCC.interruptorCerrado("matriz")){
	delete pLU;
	delete pP;
	delete pInv;
    }else{
	pCC.cierraInterruptor("matriz");
    }

    if (pCC.interruptorCerrado("vector")){
	delete pV;
    }else{
	pCC.cierraInterruptor("vector");
    }

    // Dimensin
    pN = a.filas();
    // Crea un puntero a un objeto tipo matriz para LU
    pLU = new RMatriz(pN,pN);
    // Crea un puntero a un objeto tipo matriz para inversa
    pInv = new RMatriz(pN,pN);
    // Crea un puntero a un objeto tipo permutacin
    pP = new Permutacion(pN);
    // Copia a en lu
    *pLU = a;
    // crea un puntero a un vector de dimensin N
    pV = new RVector(pN);
    // copia el vector v
    *pV = v;

    // Nueva matriz y nuevo vector en la mquina. Por tanto hay que
    // factorizar de nuevo, resolver de nuevo el sistema
    // lineal si procede y calcular de nuevo la inversa
    // y el determinante. 

    // an no hemos factorizado la matriz
    pCC.abreInterruptor("lu"); 
    // an no hemos resuelto el sistema lineal 
    pCC.abreInterruptor("lineal"); 
    // an no hemos calculado inversa
    pCC.abreInterruptor("inversa"); 
    // an no hemos calculado el determinante
    pCC.abreInterruptor("determinante"); 

}
// Establece el nivel de informacin
void ALineal::eInfo(string nivel){
    pNivel = nivel;
    if (pNivel == "detallado" ||
	pNivel == "Detallado" ||
	pNivel == "DETALLADO" ||
	pNivel == "D"){
	pNivel = "D";
    }
    else if (pNivel == "silencioso" ||
	     pNivel == "Silencioso" ||
	     pNivel == "SILENCIOSO" ||
	     pNivel == "S"){
	pNivel = "S";
    }
    else{
	cout << " nivel no vlido";
	cout << " niveles vlidos son, Silencioso o Detallado"
	     << endl;
	pNivel = "M";
    }
	
}

// Resuelve el sistema lineal
RVector ALineal::x() 
{
    // Para resolver el sistema es necesario que
    // se hayan establecido una matriz y un vector,
    // en cuyo caso los correspondientes interruptores
    // estn cerrados. Si no es as, termina con error.

    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecida",
	      CENOINI);
    if (pCC.interruptorAbierto("vector"))
	error("El vector del sistema no ha sido establecida",
	      CENOINI);
	      
    // Si el sistema ya ha sido resuelto para el
    // estado actual de la mquina el interruptor
    // correspondiente est cerrado
    
    if (pCC.interruptorCerrado("lineal")) {
	return *pV;     // vector ya calculado
    }
    else{
	// Hay que resolver de nuevo el sistema lineal
	if (pLU->filas() != pV->dimension())
	    error("LU: dime vector debe ser = a dim matriz", 
		  CEMALDIM);
	
	// Si la matriz todava no ha sido factorizada LU para el estado
	// actual de la mquina, el interruptor
	// correspondiente est abierto.
 
	if (pCC.interruptorAbierto("lu")){
	    // Descompone la matriz LU
	    // Toma en el argumento la matriz LU y la permutacin
	    int status = descomponeLU(*pLU,*pP);
	    if (status) error("Error en LU",status);

	    // La matriz ha sido factorizada 
	    // cierra el correspondiente interruptor
	    pCC.cierraInterruptor("lu"); 
	}

	// resuelve el sistema
	// toma en el argumento la matriz LU, la permutacin y el vector
	int status = resuelveLU(*pLU, *pP, *pV);
	if (status) error("Error en LU::resuelve",status);

	// El sistema lineal ha sido resuelto
	// cierra el correspondiente interruptor
	pCC.cierraInterruptor("lineal"); 
	return *pV;
    }
}

RMatriz ALineal::lu() 
{
    // Hay una matriz en la mquina?
    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecida",
	      CENOINI);

    // Si la matriz ha sido descompuesta LU el 
    // correspondiente interruptor est cerrado
    if (pCC.interruptorCerrado("lu")){
	return *pLU;
    }
    else {
	// Descompone la matriz LU
	// Toma en el argumento la matriz LU y la permutacin
	int status = descomponeLU(*pLU,*pP);
	if (status) error("Error en LU",status);

	// La matriz ha sido factorizada 
	// vierra el correspondiente interruptor
	pCC.cierraInterruptor("lu"); 
	return *pLU;
    }
}
// Extrae la matriz L de la matriz compacta LU
RMatriz ALineal::l()
{
    // Hay una matriz en la mquina?
    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecida",
	      CENOINI);

    // Descompone LU si an no lo est
    if (pCC.interruptorAbierto("lu")){
	int status = descomponeLU(*pLU,*pP);
	if (status) error("Error en LU",status);
	pCC.cierraInterruptor("lu"); 
    }

    // Y extrae la matriz L
    RMatriz lo(pN,pN);
    for (int i = 1; i <=pN; i++)
	for (int j= 1; j<=i; j++){
	    lo(i,j) = (*pLU)(i,j);
	    if (i == j) lo(i,i) = 1.;
	}
    return lo;
}

// Extrae la matriz U de la matriz compacta LU
RMatriz ALineal::u()
{
    // Hay una matriz en la maquina?
    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecida",
	      CENOINI);

    // Descompone LU si an no lo est
    if (pCC.interruptorAbierto("lu")){
	int status = descomponeLU(*pLU,*pP);
	if (status) error("Error en LU",status);
	pCC.cierraInterruptor("lu"); 
    }
    
    // Y extrae la matriz U
    RMatriz up(pN,pN);
    for (int i = 1; i <=pN; i++)
	for (int j= i; j<=pN; j++)
	    up(i,j) = (*pLU)(i,j);
    
    return up;
}

// Devuelve la matriz de permutaciones
Permutacion ALineal::p() const
{
    return *pP;
}

// Calcula la inversa de A resolviendo el sistema lineal
// PLU = W para los vectores de la base
RMatriz ALineal::inversa()
{
    // Hay una matriz en la mquina?
    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecida",
	      CENOINI);
    
    // Si ya hemos calculado la inversa el
    // correspondiente interruptor est cerrado
    if (pCC.interruptorCerrado("inversa")){
	return *pInv;
    }
    else {
	// Si el interruptor est abierto
	// descompone la matriz LU
    
	if (pCC.interruptorAbierto("lu")){
	    // Descompone la matriz LU
	    // Toma en el argumento la matriz LU y la permutacin
	    int status = descomponeLU(*pLU,*pP);
	    if (status) error("Error en LU",status);
	    pCC.cierraInterruptor("lu"); 
	}

	// ai es una referencia a la matriz inversa
	RMatriz&  ai = *pInv; 
	// Inicializamos a la matriz unidad
	ai.asignaIdentidad();

	// Iteramos sobre los vectores de la base
	for (int i = 1; i<= pN; i++){
	    // Proyectamos las columnas de la matriz
	    // (vectores de la base)
	    RVector x = proyectaColumna(ai, i);

	    // x es el vector de incgnitas a la salida
	    int status = resuelveLU(*pLU,*pP,x);
	    if (status) error("Error en LU::inversa",status);
	    
	    // copia x an ainv
	    copiaColumna(ai, x, i);	
	}
	
	// Hemos calculado la inversa, cierra el interruptor
	pCC.cierraInterruptor("inversa");
	return ai;
    }
}

//  calcula el determinante de A
double ALineal::det()
{
    // Hay una matriz en la mquina?
    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecida",
	      CENOINI);

    // Si ya hemos calculado el determinante el
    // correspondiente interruptor est cerrado
    if (pCC.interruptorCerrado("determinante")){
	return pDet;
    }
    else {
	// Si el interruptor est abierto
	// descompone la matriz LU
	
	if (pCC.interruptorAbierto("lu")){
	    int status = descomponeLU(*pLU,*pP);
	    if (status) error("Error en LU",status);

	    pCC.cierraInterruptor("lu"); 
	}

	// Accede a la matriz u (superior) mediante
	// una referencia constante (para simplificar
	// notacin)
	const RMatriz& u = this->u();

	double det = 1;
	for (int k = 1; k <= pN; k++){
	    det *=u(k,k);
	}
	// El signo del determinante es la paridad de la permutacin
	pDet = det*pP->signo();

	// cierra el interruptor
	pCC.cierraInterruptor("determinante");
	return pDet;
    }
}

/* Factorizacin de una matriz general A(N x N) en la forma:
 *
 *   P A = L U
 *
 * donde P es una matriz de permutaciones, L es triangular inferior y U
 * es triangular superior.
 * 
 * P se guarda en la permutacin p. Si el elemento \(i\) de la permutacin
 * toma el valor \(k\) entonces \(p\bp{ik}=1\), el resto de los elementos
 * de la matriz son nulos. El signo de la permutacin es \((-1)^n\),
 * donde \(n\) es el nmero de intercambios en la permutacin.
*/

int ALineal::descomponeLU(RMatriz& lu,Permutacion& p)
{

    // uij matriz triangular superior
    // lij matriz triangular inferior
    // 
    //      | u11 u12 u13...u1n |
    //      | 0   u22 u23...u2n |
    //uij=  | 0   0   u33...u3n |
    //      | 0   0   0  ...0   |
    // 
    //      | 1    0    0... 0 |
    //      | l21  1    0... 0 |
    //lij=  | l31 l32   0... 0 | 
    //      | ...
    //      | ln1 ln2  ln3...1 |

    // Itera sobre cada columna de la matriz

    int pN = lu.filas();

    if(pNivel == "D")
	cout << " Comienza descomposicin LU " << endl;

    for (int j = 1; j <= pN; j++){
	char c;
	if(pNivel == "D"){
	    cout << " Matriz m = " << lu <<endl;
	    cout << " Pulse una tecla para continuar " << endl;
	    cin >> c;
	    cout << "columna j="<<j << endl;
	    cout << "Pre-calcula uij y lij" <<endl;
	}

	// Clculo de factores
	// uij --- 
	for (int i = 1; i <= j; i++){
	    lu(i,j) = lu(i,j) - sum(i,j,1,i-1,lu);
	    if(pNivel == "D")
		cout << "u"<<i<<j<<" = "<<lu(i,j)<<endl;
	    
	}
	// lij --- matriz triangular inferior
	for (int i = j+1; i <= pN; i++){
	    lu(i,j) = lu(i,j) - sum(i,j,1,j-1,lu);
	    if(pNivel == "D")
		cout << "l"<<i<<j<<" = "<<lu(i,j)<<endl;
	}

	if(pNivel == "D"){
	    cout << " Pulse una tecla para continuar " << endl;
	    cin >> c;
	}

	// Buscamos el pivote como el elemento mayor del 
	// conjunto (ujj,lj+1,lj+2,...lnj)

	// Crea un vector de dimensin n que contiene una copia
	// de la columna j de la matriz
	RVector vc = proyectaColumna(lu, j);

	// Encuentra el mximo del vector a partir de la diagonal:
	int iPivote = maximo(vc,j);
	
	// Si el pivote es cero el mtodo fracasa
	
	if (vc(iPivote) ==0){
	    error("Imposible factorizar LU, columna de ceros",
		  CEFACTOR);
	    return CEFACTOR;
	}
	if(pNivel == "D")
	    cout << "El elemento mximo es "<< vc(iPivote) << endl;

	// Si el ndice del pivote es mayor que la diagonal
	// hay que permutar 

	if (iPivote > j){

	    if(pNivel == "D"){
		cout << "Intercambia fila "<< iPivote 
		     << " y fila " <<j<< endl;
	    }
	    lu.intercambiaFilas(j,iPivote);
	    p.intercambia(j,iPivote);
	}
	else
	    if(pNivel == "D")
		cout << "No es necesario pivotar " << endl;

	if(pNivel == "D"){
	    cout << " Pulse una tecla para continuar " << endl;
	    cin >> c;
	}

	// Divide por el pivote
	for (int i = j+1; i <= pN; i++){
	    lu(i,j) = lu(i,j)/lu(j,j);
	    if(pNivel == "D")
		cout << "l"<<i<<j<<" = "<<lu(i,j)<<endl;
	}
	if(pNivel == "D"){
	    cout << " Pulse una tecla para continuar " << endl;
	    cin >> c;
	}
    }

    return EXITO;
}

// resuelve el sistema lineal LU*X = PW
// Donde P es la matriz de permutaciones obtenida
// cuando se descompone la matriz original para obtener LU
// Devuelve el vector de incgnitas sustituyendo al de coeficientes

int ALineal::resuelveLU(const RMatriz& lu, const Permutacion& p,
			RVector& x)
{
    
    int pN = lu.filas();

    // La ecuacin a resolver es la siguiente:
    // AX = PW --> (LU)X = PW --> L(UX) = PW
    // UX = Y
    // LY = PW
    
    // Primero resolvemos LY = PW donde
    // P es la matriz de permutaciones
    // L la matriz inferior
    // W es el vector de soluciones
    // Y reemplaza a W
    // \(y\sbi = w\sbi - \Sum\sb{j=1}\sp{i-1} l\sb{ij}y\sbj\)

    x = p*x;  // LY = PW

    for (int i= 1; i<=pN; i++){
	for (int j = 1; j <= i-1; j++){
	    x(i) = x(i) - lu(i,j)*x(j);
	}
    }
    
    // A continuacin resolvemos
    // UX = Y
    // \(x\sbi = y\sbi - \Sum\sb{j=1+1}\sp{n} u\sb{ij}x\sbj\)
    // \(x\sbi = x\sbi/u\sb{ii}\)
    
    for (int i= pN; i>=1; i--){
	for (int j = i+1; j <= pN; j++){
	    x(i) = x(i) - lu(i,j)*x(j);
	}
	x(i) = x(i)/lu(i,i);
    }
    
    return EXITO;
}
