/* -*- mode: c++ -*- */
// rdiagonal.h
// Mquina para diagonalizar matrices reales y simtricas
// Implementacin de la clase RDiagonal
#include <rdiagonal.h>
#include <error.h>
#include <matutil.h>    

// Constructor
// Establece los interruptores necesarios para el
// funcionamiento del algoritmo
RDiagonal::RDiagonal()
{

    // Aade interruptores (abiertos por defecto)

    // an no hemos establecido la matriz
    pCC.nuevoInterruptor("matriz"); 

    // an no hemos diagonalizado la matriz
    pCC.nuevoInterruptor("diagonal"); 

}

// Destructor
RDiagonal::~RDiagonal()
{
    // si se crearon objetos de tipo matriz borralos
    if (pCC.interruptorCerrado("matriz")){
	delete pM;
	delete pD;
	delete pV;
    }
}

void RDiagonal::eMatriz(const SMatriz& m, 
			double precision, int maxiter)
{

    // Borra la antigua estructura
    if (pCC.interruptorCerrado("matriz")){
	delete pM;
	delete pD;
	delete pV;
    }
    else{
	pCC.cierraInterruptor("matriz");
    }

    // Puntero a la matriz simtrica
    pM = new SMatriz(m);
    // Puntero a la matriz diagonalizada
    pD = new RMatriz(m.rmatriz());
    // Puntero a la matriz de autovectores
    pV = new RMatriz(m.rmatriz());
    pEps = precision;
    pMaxiter = maxiter;

    // Nueva matriz en la mquina. Por tanto hay que
    // diagonalizar de nuevo.

    // an no hemos diagonalizado la matriz
    pCC.abreInterruptor("diagonal"); 

}

// Funciones de acceso
// devuelve la matriz
SMatriz& RDiagonal::matriz() 
{
    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecidad",
	      CENOINI);

    return *pM;
}    
// devuelve matriz diagonal
RMatriz& RDiagonal::diagonal()
{
    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecidad",
	      CENOINI);

    if (pCC.interruptorAbierto("diagonal")){
	// diagonaliza la matriz
	jacobi(*pD, *pV, pEps, pMaxiter);
	// cierra el interruptor
	pCC.cierraInterruptor("diagonal"); 
    }

    return *pD;
}  
// devuelve matriz autovectores
RMatriz& RDiagonal::autovectores()   
{
    if (pCC.interruptorAbierto("matriz"))
	error("La matriz del sistema no ha sido establecidad",
	      CENOINI);

    if (pCC.interruptorAbierto("diagonal")){
	// diagonaliza la matriz
	jacobi(*pD, *pV, pEps, pMaxiter);
	pCC.cierraInterruptor("diagonal"); 
    }

    return *pV;
}

// Implementacin del algoritmo de Jacobi para
// diagonalizar matrices reales y simtricas

// La matriz es a la entrada una matriz simtrica
// En cada iteracin contiene la aproximacin a 
// los autovalores
// La matriz r contiene aproximaciones sucesivas a los autovectores

void RDiagonal::jacobi(RMatriz& a, RMatriz& r, double eps, double maxiter)
{
    // Inicializamos r a la matriz unidad
    r.asignaIdentidad();

    // Iteramos hasta alcanzar el nmero mximo
    // de iteraciones o la precisin requerida
    // en el elemento mayor fuera de la diagonal mayor

    int iter =0;   // iteraciones
    double max = BIG; // elemento mximo

    while (iter <= maxiter && abs(max) > eps){
	int p,q;
	// Encontramos el elemento mximo fuera de la diagnonal
	max = maxNoDiag(a,p, q);
	if (max == 0) break; // matriz ya diagonal
	// Efectuamos una rotacin de jacobi en dicho elemento
	rotJacobi(a,r,p,q);
	iter++;
    }
}

void  RDiagonal::rotJacobi(RMatriz& a, RMatriz& r, int p, int q)	
{
    // Efecta una rotacin teta de Jacobi en el plano pq, tal que
    // el trmino a(p,q)=0. 
    // La rotacin se almacena en r, cuyas columnas sern los autovectores
    // de a al finalizar el proceso
    // n es la dimensin de la matriz a (y r)

    int n = a.filas();  // matriz simtrica, filas = columnas
    // d = cot(2teta) frmula 1.12 
    double d = (a(q,q)-a(p,p))/(2.0*a(p,q));   
    double t;
    if (abs(d) < BIG){
	//t = tan(teta) frmula 1.14
	t = signo(d)/(fabs(d) + sqrt(d*d + 1)); 
    }
    else{
	t = 1/(2*d);
    }

    // c = cos(teta)
    double c = 1.0/sqrt(t*t + 1.0); 
    // s = sin(teta)
    double s = t*c;  
    double tau = s/(1+c);

    // Ecuaciones (1.17)
    
    a(p,p) -= t*a(p,q);
    a(q,q) += t*a(p,q);
    a(p,q) = a(q,p) = 0.0;

    for (int i = 1; i <= n; i++){
	if (i != p && i!= q){
	    double aip = a(i,p) - s*(a(i,q) + tau*a(i,p));
	    double aiq = a(i,q) + s*(a(i,p) - tau*a(i,q));
	    a(i,p) = a(p,i) = aip;
	    a(i,q) = a(q,i) = aiq;

	}
    }

    // Guardamos la rotacin en la matriz r

    for (int i = 1; i <= n; i++){
	double Rip =  c*r(i,p) - s*r(i,q);
	double Riq = s*r(i,p) + c*r(i,q);
	r(i,p) = Rip;
	r(i,q) = Riq;
    }

}
